using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;

// FluidEnviroEditorRain v1.0
//
// Allows you to edit some parameters of fluid particles

namespace FluidEnviroEditorRainPlugin
{
    public class FluidEnviroEditorRain : MVRScript
    {
		
		public JSONStorableStringChooser FluidAppareance;
		protected List<string> appearanceChoices;
		
		public JSONStorableColor FluidMainColor;
		public JSONStorableFloat FluidMainOpacity;

		public JSONStorableFloat FluidAmount;
		public JSONStorableFloat FluidFallSpeedMin;
		public JSONStorableFloat FluidFallSpeedMax;
		public JSONStorableFloat FluidWindSpeedMin;
		public JSONStorableFloat FluidWindSpeedMax;

		public JSONStorableFloat FluidAreaWidth;
		public JSONStorableFloat FluidAreaLength;
		
		public JSONStorableFloat FluidLifetimeMin;
		public JSONStorableFloat FluidLifetimeMax;
		
		public JSONStorableBool EnableSound;
		public JSONStorableFloat FluidSoundVolume;
		public JSONStorableStringChooser FluidSound;
		protected List<string> soundChoices;
		
		public JSONStorableBool ShowFluidBounds;
			
		public Transform ParticleRoot;
		public Transform SfxRoot;
		private ParticleSystem prtComponent;
		private ParticleSystemRenderer prtRendererComponent;
		private AudioSource prtSoundComponent;
			
		private string objectSearchName = "hzm_fluidenviro_";

		private bool seInitialized = false;
		
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
		private GameObject debugParent;
		private GameObject boundsCube;
		
		public override void Init()
        {			 
            try
            {
				if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add FluidEnviroEditorRain on an Custom Unity Asset with a compatible particle system.");
                    return;
                }
				
				// Type of appearance (lighting)
				appearanceChoices = new List<string>();
				appearanceChoices.Add("Unlit");
				appearanceChoices.Add("Lit");
				
				FluidAppareance = new JSONStorableStringChooser("Rain appearance", appearanceChoices, "Unlit", "Rain appearance", FluidAppareanceCallback);
				UIDynamicPopup FAdp = CreatePopup(FluidAppareance, false);
				FAdp.labelWidth = 300f;
				
				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				FluidMainColor = new JSONStorableColor("Rain color", hsvc, MainColorCallback);
				CreateColorPicker(FluidMainColor);
				
				FluidMainOpacity = new JSONStorableFloat("Rain opacity", 0.20f, MainOpacityCallback, 0f, 1f);
				CreateSlider(FluidMainOpacity, false);

				FluidAmount = new JSONStorableFloat("Rain amount", 1500f, FluidAmountCallback, 100f, 28000f);
				CreateSlider(FluidAmount, false);
				
				FluidFallSpeedMin = new JSONStorableFloat("Rain fall speed min", 13f, FluidFallSpeedMinCallback, 1f, 50f);
				CreateSlider(FluidFallSpeedMin, false);
				
				FluidFallSpeedMax = new JSONStorableFloat("Rain fall speed max", 15f, FluidFallSpeedMaxCallback, 1f, 50f);
				CreateSlider(FluidFallSpeedMax, false);
				
				FluidWindSpeedMin = new JSONStorableFloat("Rain wind speed min", 0f, FluidWindSpeedMinCallback, -10f, 10f);
				CreateSlider(FluidWindSpeedMin, false);
				
				FluidWindSpeedMax = new JSONStorableFloat("Rain wind speed max", 0f, FluidWindSpeedMaxCallback, -10f, 10f);
				CreateSlider(FluidWindSpeedMax, false);
				
				FluidAreaWidth = new JSONStorableFloat("Rain area width", 20f, FluidAreaWidthCallback, 1f, 50f);
				CreateSlider(FluidAreaWidth, false);
				
				FluidAreaLength = new JSONStorableFloat("Rain area length", 20f, FluidAreaLengthCallback, 1f, 50f);
				CreateSlider(FluidAreaLength, false);

				FluidLifetimeMin = new JSONStorableFloat("Rain lifetime (min)", 1.7f, FluidLifetimeMinCallback, 0.5f, 3f);
				CreateSlider(FluidLifetimeMin, false);
				
				FluidLifetimeMax = new JSONStorableFloat("Rain lifetime (max)", 1.7f, FluidLifetimeMaxCallback, 0.5f, 3f);
				CreateSlider(FluidLifetimeMax, false);				

				EnableSound = new JSONStorableBool("Enable Sound", false, EnableSoundCallback);
				CreateToggle(EnableSound, false);
				
				FluidSoundVolume = new JSONStorableFloat("Sound volume", 1f, FluidSoundVolumeCallback, 0.01f, 1f);
				CreateSlider(FluidSoundVolume, false);
				
				// Type of appearance (lighting)
				soundChoices = new List<string>();
				soundChoices.Add("Rain city heavy");
				soundChoices.Add("Rain city backyard light");
				
				FluidSound = new JSONStorableStringChooser("Sound", soundChoices, "Rain city heavy", "Sound", FluidSoundCallback);
				UIDynamicPopup FSdp = CreatePopup(FluidSound, false);
				FSdp.labelWidth = 110f;
				
				ShowFluidBounds = new JSONStorableBool("Show bounds", false, ShowBoundsCallback);
				CreateToggle(ShowFluidBounds, false);
				
				CreateSpacer();
				CreateSpacer();
				CreateSpacer();
				
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Rain</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>IMPORTANT:</b> If you switch particles in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>Performances:</b> The particle system does not optimize itself. You are free to push everything to the max, but framerate will potentially drop. Some tips and tricks are available in the documentation on the hub.\n\n" +
					"<b>Rain appearance</b> is the way the rain looks depending on the light.\n" +
					"<b><i>- Unlit</i></b> ignores the light which means the fluid will be visible in complete darkness.\n" +
					"<b><i>- Lit</i></b> takes in account the lighting, which means the fluid will be invisible in complete darkness but it will also be coloured by the lights around.\n\n" +
					"<b>Sound:</b> Enables rain sound. These sound are global (you hear them no matter where you are), avoid using sound on each rain CUA, one is enough. More info is available in the documentation on the hub.\n\n" +
					"<b>Show Bounds:</b> is a visual gizmo to help you position the rain. It is not saved with the scene and will be disabled on scene load.\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1080.0f;
								
				RegisterStringChooser(FluidAppareance);
				RegisterColor(FluidMainColor);
				RegisterFloat(FluidMainOpacity);
				RegisterFloat(FluidAmount);
				RegisterFloat(FluidFallSpeedMin);
				RegisterFloat(FluidFallSpeedMax);
				RegisterFloat(FluidWindSpeedMin);
				RegisterFloat(FluidWindSpeedMax);
				RegisterFloat(FluidAreaWidth);
				RegisterFloat(FluidAreaLength);
				RegisterFloat(FluidLifetimeMin);
				RegisterFloat(FluidLifetimeMax);
				RegisterBool(EnableSound);
				RegisterFloat(FluidSoundVolume);
				RegisterStringChooser(FluidSound);
				
				// ************************************************
				// ** ACTIONS
				// ************************************************			


			}
            catch(Exception e)
            {
                SuperController.LogError("FluidEnviroEditorRain - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			
        }
		
		void Update()
        {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && seInitialized == false ) {
				initLoops++;
				InitDE();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("FluidEnviroEditorRain : This CUA particle could not be initialized properly after 3600 tries. Please select a compatible particle and reload the script.");
				initLoops++; // To avoid spamming the log
			}
		}
		
		void OnDisable() {
			if( debugParent != null ) GameObject.Destroy(debugParent);
		}
		
		private bool InitDE() {
			if( seInitialized == true ) return true;
			
			ParticleRoot = getObjectRoot( containingAtom.reParentObject );
			if( ParticleRoot == null ) {
				return false;
			}
			
			SfxRoot = getSfxRoot( ParticleRoot );
			if( SfxRoot == null ) {
				return false;
			}
			prtSoundComponent = SfxRoot.gameObject.GetComponent<AudioSource>();

			// Getting motion mode child
			foreach( Transform child in ParticleRoot )
			{
				
				child.gameObject.SetActive(true);
				prtComponent = child.GetComponent<ParticleSystem>();
				if( prtComponent ) {
					prtRendererComponent = child.GetComponent<ParticleSystemRenderer>();
					
					seInitialized = true;
					
					UpdateParticleSettings(); // Sets all the parameters of the particle
					ChangeFluidChildParticle( FluidAppareance.val ); // Change the particle appearance
					
					logDebug("FluidEnviroEditorRain : Fluid editor initialized properly");

					return true;
				} else {
					logError("FluidEnviroEditorRain : your CUA is incorrect and cannot be used with this script.");
					initLoops = maxInitLoops; // We don't have a particle system, we consider the initialization failed
					return false;
				}
				break;
			}
			
			return false;

		}
		
		private Transform getObjectRoot( Transform parent ) {
			foreach( Transform child in parent )
			{
				if( child.name.Contains(objectSearchName) ) {
					return child;
				} else {
					Transform childSearch = getObjectRoot(child);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}
		
		private Transform getSfxRoot( Transform parent ) {
			foreach( Transform child in parent )
			{
				if( child.name == "fluid_sfx" ) {
					return child;
				} else {
					Transform childSearch = getSfxRoot(child);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}
		
		private void UpdateParticleSettings() {
			MainColorCallback( FluidMainColor );
			MainOpacityCallback( FluidMainOpacity.val );
			
			FluidAmountCallback( FluidAmount.val );
			FluidFallSpeedMinCallback( FluidFallSpeedMin.val );
			FluidWindSpeedMinCallback( FluidWindSpeedMin.val );

			FluidAreaWidthCallback(FluidAreaWidth.val);
			FluidLifetimeMinCallback( FluidLifetimeMin.val );
			
			FluidSoundCallback( FluidSound.val );
		}
		
		protected void ChangeFluidChildParticle( string particleName ) {
			if( ParticleRoot != null ) {
							
				int loopCount = 0;
				foreach( Transform child in ParticleRoot )
				{
					if( child.name != "fluid_sfx" ) child.gameObject.SetActive(false);
					if( loopCount == appearanceChoices.IndexOf(particleName) ) {
						child.gameObject.SetActive(true);
						prtComponent = child.GetComponent<ParticleSystem>();
						if( prtComponent ) {
							prtRendererComponent = child.GetComponent<ParticleSystemRenderer>();
							
						}
					}
					
					loopCount++;
				}
			
				// Setting parameters
				UpdateParticleSettings();
			}
		}
		
		protected void UpdateBoundsCube() {
			if ( boundsCube != null ) {
				boundsCube.transform.localScale = new Vector3( FluidAreaWidth.val, 0.2f, FluidAreaLength.val );
			}
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		protected void FluidAppareanceCallback(string appearanceChoice) {
			ChangeFluidChildParticle( appearanceChoice );
		}
				
		protected void MainColorCallback(JSONStorableColor color) {
			if( prtComponent ) {
				Color newColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, FluidMainOpacity.val);
				var prtMain = prtComponent.main;
				prtMain.startColor = newColor;
			}
		}
		
		protected void MainOpacityCallback(float opacity) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startColor = new Color(FluidMainColor.colorPicker.currentColor.r, FluidMainColor.colorPicker.currentColor.g, FluidMainColor.colorPicker.currentColor.b, opacity);
			}
		}
		
		protected void FluidAmountCallback(float amount) {
			if( prtComponent ) {
				var prtEmission = prtComponent.emission;
				prtEmission.rateOverTime = amount;
			}
		}
		
		protected void FluidWindSpeedMinCallback(float speed) {
			if( prtComponent ) {
				var prtVOL = prtComponent.velocityOverLifetime;
				prtVOL.x = new ParticleSystem.MinMaxCurve(speed, FluidWindSpeedMax.val);
			}
		}
		
		protected void FluidWindSpeedMaxCallback(float speed) {
			if( prtComponent ) {
				var prtVOL = prtComponent.velocityOverLifetime;
				prtVOL.x = new ParticleSystem.MinMaxCurve(FluidWindSpeedMin.val, speed);
			}
		}

		protected void FluidFallSpeedMinCallback(float speed) {
			if( prtComponent ) {
				var prtVOL = prtComponent.velocityOverLifetime;
				prtVOL.y = new ParticleSystem.MinMaxCurve(-speed, -FluidFallSpeedMax.val);
			}
		}
		
		protected void FluidFallSpeedMaxCallback(float speed) {
			if( prtComponent ) {
				var prtVOL = prtComponent.velocityOverLifetime;
				prtVOL.y = new ParticleSystem.MinMaxCurve(-FluidFallSpeedMin.val, -speed);
			}
		}
		
		protected void FluidAreaWidthCallback(float width) {
			if( prtComponent ) {
				var prtShape = prtComponent.shape;
				prtShape.scale = new Vector3(width,0.5f,FluidAreaLength.val);
				
				UpdateBoundsCube();
			}
		}
		
		protected void FluidAreaLengthCallback(float depth) {
			if( prtComponent ) {
				var prtShape = prtComponent.shape;
				prtShape.scale = new Vector3(FluidAreaWidth.val,0.5f,depth);
				
				UpdateBoundsCube();
			}
		}
		
		protected void FluidLifetimeMinCallback(float lifetimeMin) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startLifetime = new ParticleSystem.MinMaxCurve(lifetimeMin, FluidLifetimeMax.val);
			}
		}
		
		protected void FluidLifetimeMaxCallback(float lifetimeMax) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startLifetime = new ParticleSystem.MinMaxCurve(FluidLifetimeMin.val, lifetimeMax);
			}
		}
		
		protected void EnableSoundCallback(bool enabled) {
			if( prtComponent ) {
				FluidSoundCallback(FluidSound.val);
			}
		}
		
		protected void FluidSoundVolumeCallback(float sndVolume) {
			if( prtSoundComponent ) {
				prtSoundComponent.volume = sndVolume;
			}
		}
		
		protected void FluidSoundCallback(string sndName) {
			if( prtComponent && SfxRoot ) {
				foreach( Transform child in SfxRoot ) {
					child.gameObject.SetActive(false);
					if( EnableSound.val == true && child.name == sndName ) {
						child.gameObject.SetActive(true);
						prtSoundComponent = child.gameObject.GetComponent<AudioSource>();
						FluidSoundVolumeCallback( FluidSoundVolume.val );
					}
				}
			}
		}
		
		protected void ShowBoundsCallback(bool enabled) {
			if( prtComponent ) {
				if( debugParent == null ) {
					debugParent = new GameObject();
					debugParent.name = "hzmDebugParent";
					debugParent.transform.SetParent(ParticleRoot.parent.transform, false);
					debugParent.transform.rotation = ParticleRoot.parent.transform.rotation;
				}
				
				if( enabled == true ) {
					boundsCube = GameObject.CreatePrimitive(PrimitiveType.Cube);
					boundsCube.name = "hzmBoundsCube";
					boundsCube.transform.SetParent(debugParent.transform,false);
					boundsCube.transform.rotation = debugParent.transform.rotation;
					
					MeshRenderer mr = boundsCube.GetComponent<MeshRenderer>();
					mr.material = new Material(Shader.Find("Particles/Additive"));
					mr.material.SetColor("_TintColor", new Color(0.8f, 0.1f, 1f, 0.01f));
					
					UpdateBoundsCube();
				} else {
					GameObject.Destroy(debugParent);
					debugParent = null;
					boundsCube = null;
				}
			}
		}
		

		// **************************
		// COROUTINES
		// **************************

		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}


	}
}